; ==============================================================================
; Keyboard controller functions
; Copyright 2004 Emboss, all rights reserved
; See licence.txt for details.
; ==============================================================================

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdEmptyBuf
;   Empties the input and output buffers of the keyboard controller.
;   Sets ZF on success, clears ZF on fail.
;   Destroys al and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdEmptyBuf:
  xor     cx,cx
.EmptyLoop:
  in      al,0x64                   ; Read in the keyboard status
  IODELAY                           ;   bit 0 = set if output byte available
  and     al,3                      ;   bit 1 = set if input buffer full
  jz      .Done
  and     al,1
  jz      .NextTry
  in      al,0x60                   ; Read in the waiting byte and discard
  IODELAY
.NextTry:
  dec     cx
  jnz     .EmptyLoop
  or      cl,1                      ; clears ZF
.Done:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdInBufWait
;   Waits for the input (ie: computer to keyboard) buffer to become empty.
;   Sets ZF on success, clears ZF on fail.
;   Destroys al and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdInBufWait:
  xor     cx,cx
.WaitLoop:
  in      al,0x64                   ; Read in the keyboard status
  IODELAY
  and     al,2                      ; bit 1 set if input buffer full
  jz      .BufferReady
  dec     cx
  jnz     .WaitLoop
  or      cl,1                      ; Clears ZF
.BufferReady:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdOutBufWait
;   Waits for the output (ie: keyboard to computer) buffer to become empty.
;   Sets ZF on success, clears ZF on fail.
;   Destroys al and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdOutBufWait:
  xor     cx,cx
.WaitLoop:
  in      al,0x64                   ; Read in the keyboard status
  IODELAY
  and     al,1                      ; bit 0 set if output byte available
  jnz     .BufferReady
  dec     cx
  jnz     .WaitLoop
  or      cl,1                      ; Clears ZF
  jmp     .Exit
.BufferReady:
  xor     cl,cl
.Exit:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdRead
;   Reads a byte from the keyboard through port 0x60.
;   Sets ZF on success, clears ZF on fail. The byte read (if any) is in al.
;   Destroys al and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdRead:
  call    KbdOutBufWait             ; Wait for some data to arrive
  jnz     .Exit
  in      al,0x60                   ; Read the data (does not affect ZF)
  IODELAY
.Exit:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdWrite
;   Writes a byte to the keyboard through port 0x60.
;   Sets ZF on success, clears ZF on fail. The byte read (if any) is in al.
;   Destroys ax and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdWrite:
  mov     ah,al
  call    KbdInBufWait              ; Wait for some data to arrive
  jnz     .Exit
  mov     al,ah
  out     0x60,al                   ; Write the data (does not affect ZF)
  IODELAY
.Exit:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdSendCmd60
;   Sends a command through port 60 of the keyboard controller. Checks to make
;   sure that it is ack'd, and resends if requested.
;   The command to send is in AL
;   Sets ZF on success, clears ZF on fail.
;   Destroys ax, bx and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdSendCmd60:
  mov     bl,10                     ; Do up to 10 retries
  mov     bh,al
.ResendLoop:
  mov     al,bh
  call    KbdWrite
  jnz     .Exit
  mov     al,ah
  call    KbdRead                   ; Get the result
  jnz     .Exit
  cmp     al,0xFA
  je      .Exit                     ; Jump if ACK (ZF set)
  cmp     al,0xFE
  jne     .Fail                     ; Jump if not RESEND
  dec     bl
  jnz     .ResendLoop               ; Loop around if maximum resends not reached
.Fail:
  or      bl,1                      ; clear ZF
.Exit:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdSendCmd64
;   Sends a command through port 64 of the keyboard controller.
;   The command to send is in AL
;   Sets ZF on success, clears ZF on fail.
;   Destroys ax, bl and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdSendCmd64:
  mov     ah,al
  call    KbdInBufWait              ; Wait to be cleared to send
  jnz     .Fail
  mov     al,ah                     ; ZF not affected
  out     0x64,al                   ; Send the command (ZF not affected)
  IODELAY
.Fail:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdSendCmd64_R
;   Sends a command through port 64 of the keyboard controller, then reads a
;   byte from port 0x60.
;   The command to send is in AL
;   Sets ZF on success, clears ZF on fail. On success AL contains the byte read.
;   Destroys ax, bl and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdSendCmd64_R:
  call    KbdSendCmd64
  jnz     .Exit
  call    KbdRead
.Exit:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; KbdInit
;   Initialises the keyboard.
;   Sets ZF on success, clears ZF on fail.
;   Destroys ax, bl and cx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
KbdInit:
  call    KbdEmptyBuf               ; Empty the buffer
  jnz     .Exit

  mov     al,0xAA                   ; Send a controller self-test request
  call    KbdSendCmd64_R
  jnz     .Exit
  cmp     al,0x55
  jnz     .Exit

  mov     al,0xAB                   ; Send an interface self-test request
  call    KbdSendCmd64_R
  jnz     .Exit
  or      al,al
  jnz     .Exit

  mov     al,0xAE                   ; Enable the keyboard
  call    KbdSendCmd64
  jnz     .Exit
  mov     al,0xFF                   ; Reset the keyboard
  call    KbdSendCmd60
  jnz     .Exit
  call    KbdRead                   ; Wait for power-on reset byte
  jnz     .Exit
  cmp     al,0xAA
  jnz     .Exit

  mov     al,0xAE                   ; Disable the keyboard
  call    KbdSendCmd64
  jnz     .Exit
  mov     al,0x60                   ; Set the controller command byte
  call    KbdSendCmd64
  jnz     .Exit
  mov     al,0x44                   ; no interrupts, but otherwise normal
  call    KbdWrite
  jnz     .Exit
  mov     al,0xF4                   ; Enable the keyboard
  call    KbdSendCmd60

.Exit:
  ret                               ; All done!
