; ==============================================================================
; A basic UART driver
; Copyright 2004 Emboss, all rights reserved
; See licence.txt for details.
; ==============================================================================

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; TTYSProbe
;   Tests to see if a serial port is really there.
;   Called with bp = base port
;   Return ZF set if UART exists, ZF cleared otherwise
;   Destroys ax, dx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TTYSProbe:
  mov     dx,bp                     ; Set DLAB=1
  add     dx,3
  in      al,dx
  or      al,0x80
  out     dx,al

  mov     dx,bp                     ; Set divisor = 0x5A
  mov     al,0x5A
  out     dx,al
  in      al,dx
  cmp     al,0x5A                   ; Check to make sure it stuck
  jne     .NoPort

  mov     al,0xA5                   ; Set divisor = 0xA5
  out     dx,al
  in      al,dx
  cmp     al,0xA5                   ; Check to make sure it stuck
  jne     .NoPort

  xor     al,al
.NoPort:
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; TTYSInit
;   Initialises the serial port.
;   Called with bp = base port
;   Returns nothing
;   Destroys ax, dx
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TTYSInit:
  push    ds
  push    si

  mov     si,TTYInitSeq             ; The location of the register table
  mov     ah,0
.RegLoop:
  mov     al,[si]                   ; Get the register ...
  mov     dx,bp
  cmp     al,0xff                   ; (0xFF = end of table)
  je      NEAR .EndOfReg
  add     dx,ax
  mov     al,[si+1]                 ; ... and the value ...
  add     si,2

  out     dx,al                     ; and write it
  jmp     .RegLoop

.EndOfReg:
  pop     si
  pop     ds
  ret

TTYInitSeq:
  ; Register, Value
  db  0x03, 0x83                    ; DLAB=1, no parity, 8 data bit, 1 stop bit
  db  0x00, 0x01                    ; Low divisor = 1
  db  0x01, 0x00                    ; High divisor = 0
  db  0x03, 0x03                    ; DLAB=0, no parity, 8 data bit, 1 stop bit
  db  0x01, 0x00                    ; Disable all interupts
  db  0x02, 0x00                    ; Disable FIFO
  db  0x04, 0x00                    ; DTR and RTS off
  db  0xFF

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; TTYSReadB
;   Reads a byte from the serial port
;   Called with dx = base port
;   Returns the read byte in al
;   Destroys nothing
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TTYSReadB:
  add     dx,5
.InputPoll:
  in      al,dx
  test    al,1
  jz      .InputPoll                ; Loop around if byte not available
  sub     dx,5
  in      al,dx
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; TTYSReadW
;   Reads a word from the serial port
;   Called with dx = base port
;   Returns the read word in ax
;   Destroys nothing
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TTYSReadW:
  call    TTYSReadB
  mov     ah,al
  call    TTYSReadB
  xchg    al,ah
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; TTYSWriteB
;   Writes a byte to the serial port
;   Called with dx = base port, al = byte to write
;   Returns nothing
;   Destroys ax
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TTYSWriteB:
  mov     ah,al
  add     dx,5
.InputPoll:
  in      al,dx
  test    al,0x20
  jz      .InputPoll                ; Loop around if buffer not free
  mov     al,ah
  sub     dx,5
  out     dx,al
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; TTYSWriteW
;   Writes a word to the serial port
;   Called with dx = base port, al = byte to write
;   Returns nothing
;   Destroys ax, bl
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TTYSWriteW:
  mov     bl,ah
  call    TTYSWriteB
  mov     al,bl
  call    TTYSWriteB
  ret
