; ==============================================================================
; Serial BIOS Upload functions
; Copyright 2004 Emboss, all rights reserved
; See licence.txt for details.
; ==============================================================================

%define ACK_CHAR 'K'
%define ERR_CHAR '!'
; Version 1.0
%define SBUVERSION 0x10

Ports:
  dw  0x3F8
  dw  0x2F8
  dw  0x3E8
  dw  0x2E8
  dw  0xFFFF

mod_SBU:
  push    bp
  push    si

  mov     si,Ports                  ; Scan for a serial port
.PortLoop:
  mov     bp,[si]
  cmp     bp,0xFFFF
  je      .ExitBreak
  call    TTYSProbe
  jz      .PortFound
  add     si,2
  jmp     .PortLoop

.PortFound:
  call    TTYSInit
  jnz     .ExitBreak                ; Serial port not working, bail out
  call    TTYSProcessSBU
.Exit:
  pop     si
  pop     bp
  ret

.ExitBreak:
  or      al,1
  jmp     .Exit

SBUCommands:
  db  'S'
  dw  SBUDoSetRegion

  db  'G'
  dw  SBUDoGetRegion

  db  'U'
  dw  SBUDoSetSpeed

  db  'R'
  dw  SBUDoRxTest

  db  'T'
  dw  SBUDoTxTest

  db  'V'
  dw  SBUDoGetVersion

  db  'C'
  dw  SBUDoCallAddr

  ; D = "done"
  ; P = self-update
  
  db  0
  
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; TTYSProcessSBU
;   Performs a SBU conversation over the serial port.
;   Called with bp = base port
;   Destroys ax, dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
TTYSProcessSBU:
  SHOWSTR "TTYSProcessSBU"
  push    si
  push    cx

  mov     [SerialBase],bp
  mov     dx,bp                     ; Raise RTS and DTR
  add     dx,4
  in      al,dx
  or      al,3
  out     dx,al

  mov     bx,10
  SHOWSTR "Waiting for CTS"
  mov     dx,bp                     ; Wait for DSR and CTS to be raised
  add     dx,6
.BXLoop:
  xor     cx,cx
.DSRCTSDCDLoop:
  in      al,dx
  not     al
  and     al,0x30
  jz      .ReadyToGo
  dec     cx
  jnz     .DSRCTSDCDLoop
  dec     bx
  jnz     .BXLoop
  jmp     .Done

.ReadyToGo:
  SHOWSTR "Sending signature"
  mov     dx,bp                     ; Send signature "SBU"
  mov     al,'S'
  call    TTYSWriteB
  mov     al,'B'
  call    TTYSWriteB
  mov     al,'U'
  call    TTYSWriteB

.InputLoop:
  mov     dx,bp                     ; Get the next byte
  call    TTYSReadB

  cmp     al,'D'
  jz      .Done
  cmp     al,'P'
  jz      .DoSelfUpdate

  mov     si,SBUCommands            ; See which command has been sent
.CmdScan:
  mov     ah,[si]
  add     si,3
  or      ah,ah
  jz      .Error
  cmp     al,ah
  jne     .CmdScan
  call    [si-2]
  jmp     .InputLoop

.DoSelfUpdate:
  mov     WORD [FinishDest],SBUDoSelfUpdate ; Jump to the real update function
  mov     BYTE [KbdOverride],0x2E   ; Make it automatically go to the SBU code
  jmp     .Exit

.Done:
  SHOWSTR ".Done"
  mov     dx,bp                     ; Drop RTS and DTR
  add     dx,4
  in      al,dx
  and     al,0xfc
  out     dx,al

.Exit:
  or      al,1                      ; Clear ZF (=abort main loop)
  SHOWSTR ".Exit"
  pop     cx
  pop     si
  ret

.Error:
  mov     dx,bp                     ; Send error marker '!'
  mov     al,ERR_CHAR
  call    TTYSWriteB
  jmp     .Done

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoSetRegion
;   Processes the "S" command (copy from serial port to memory).
;   Destroys ax,dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoSetRegion:
  push    es
  push    cx
  push    di

  mov     dx,bp

  call    TTYSReadW                 ; Get the offset
  mov     di,ax
  call    TTYSReadW                 ; Get the segment
  mov     es,ax
  call    TTYSReadW                 ; Get the length
  mov     cx,ax

  ; Blat the data from the serial port to memory
.ReadLoop:
  add     dx,5
.InputPoll1:
  in      al,dx
  test    al,1
  jz      .InputPoll1
  sub     dx,5
  in      al,dx
  stosb
  loop    .ReadLoop

  mov     al,ACK_CHAR               ; Send ACK
  call    TTYSWriteB

  pop     di
  pop     cx
  pop     es
  xor     al,al
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoGetRegion
;   Processes the "G" command (copy from memory to serial port).
;   Destroys ax,dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoGetRegion:
  push    ds
  push    cx
  push    di

  mov     dx,bp

  call    TTYSReadW                 ; Get the offset
  mov     di,ax
  call    TTYSReadW                 ; Get the segment
  mov     ds,ax
  call    TTYSReadW                 ; Get the length
  mov     cx,ax

  ; Blat the data from memory to the serial port
.WriteLoop:
  add     dx,5
.OutputPoll:
  in      al,dx
  test    al,0x20
  jz      .OutputPoll
  lodsb
  sub     dx,5
  out     dx,al
  loop    .WriteLoop

  mov     al,ACK_CHAR               ; Send ACK
  call    TTYSWriteB

  pop     di
  pop     cx
  pop     ds
  xor     al,al
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoSetSpeed
;   Processes the "U" command (set the UART divisor).
;   Destroys ax,dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoSetSpeed:
  push    cx

  mov     dx,bp

  call    TTYSReadW                 ; Get the new divisor
  mov     cx,ax

  mov     al,ACK_CHAR               ; Send ACK
  call    TTYSWriteB

  add     dx,3                      ; Set DLAB = 1
  in      al,dx
  or      al,0x80
  out     dx,al

  sub     dx,2                      ; Set the low byte
  mov     al,ch
  out     dx,al

  dec     dx                        ; And the high byte
  mov     al,cl
  out     dx,al

  add     dx,3                      ; Set DLAB = 0
  in      al,dx
  and     al,0x7F
  out     dx,al

  pop     cx
  xor     al,al
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoRxTest
;   Processes the "R" command (do a receive test).
;   Destroys ax,dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoRxTest:
  push    cx

  mov     dx,bp

  call    TTYSReadB                 ; Get the number of bytes
  movzx   cx,al

  mov     ah,0
.TestLoop:
  add     dx,5
.InputPoll1:
  in      al,dx
  test    al,1
  jz      .InputPoll1
  sub     dx,5
  in      al,dx
  cmp     al,ah
  jnz     .TestFail
  inc     ah
  loop    .TestLoop

  mov     al,ACK_CHAR               ; Send ACK
  call    TTYSWriteB

  pop     cx
  xor     al,al
  ret

.TestFail:
  mov     al,ERR_CHAR               ; Send ERR
  call    TTYSWriteB

  pop     cx
  or      al,1
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoTxTest
;   Processes the "T" command (do a transmit test).
;   Destroys ax,dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoTxTest:
  push    cx

  mov     dx,bp

  call    TTYSReadB                 ; Get the number of bytes
  movzx   cx,al

  mov     ah,0
  ; Blat the data from memory to the serial port
.WriteLoop:
  add     dx,5
.OutputPoll:
  in      al,dx
  test    al,0x20
  jz      .OutputPoll
  mov     al,ah
  inc     ah
  sub     dx,5
  out     dx,al
  loop    .WriteLoop

  mov     al,ACK_CHAR               ; Send ACK
  call    TTYSWriteB

  pop     cx
  xor     al,al
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoGetVersion
;   Processes the "V" command (get SBU version information).
;   Destroys ax,dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoGetVersion:
  push    bx
  
  mov     dx,bp

  mov     al,SBUVERSION             ; Send version
  call    TTYSWriteB

  mov     ax,cs                     ; Send code segment
  call    TTYSWriteW                ; Destroys bl

  mov     al,ACK_CHAR               ; Send ACK
  call    TTYSWriteB

  xor     al,al
  
  pop     bx
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoCallAddr
;   Processes the "C" command (call an address).
;   Destroys ax,dx
;   Sets ZF on success, clears it on failure
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoCallAddr:
  push    cx

  mov     dx,bp

  call    TTYSReadW                 ; Get the offset
  mov     cx,ax
  call    TTYSReadW                 ; Get the segment

  push    cs                        ; Return segment
  push    .RetAddr                  ; Return offset
  push    ax                        ; Destination segment
  push    cx                        ; Destination offset
  retf
.RetAddr:
  pop     cx
  xor     al,al
  ret

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
; SBUDoSelfUpdate
;   Do a self-update of the SBU code. Not called normally: it's called when the
;   main patcher is finished.
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SBUDoSelfUpdate:
  mov     ax,cs
  mov     ds,ax
  mov     es,ax

  mov     dx,[SerialBase]           ; Retreive the base port number

  mov     al,ACK_CHAR
  call    TTYSWriteB                ; Send an ACK to start the transfer

  call    TTYSReadW                 ; Get the number of bytes
  mov     cx,ax

  mov     ah,5                      ; Set up things for the real update
  mov     di,AfterSelfHeader

  jmp     SBUSelfUpdate
